import datetime
import os
import platform
import random
import socket
import string
import time
import typing as t

import requests
from loguru import logger


def _read_clipboard() -> str:
    try:
        import pyperclip
    except ImportError:
        logger.error(
            "pyperclip is not installed. Please install it with `pip install pyperclip` or `pip install nerve-adk[clipboard]`"
        )
        return ""

    val = pyperclip.paste()

    logger.info(f"📋 copied {len(val)} characters from clipboard")
    logger.debug(f"clipboard content: {val}")

    return val


_builtins: dict[str, t.Callable[[], str]] = {
    # date and time
    "CURRENT_DATE": lambda: datetime.datetime.now().strftime("%Y-%m-%d"),
    "CURRENT_TIME": lambda: datetime.datetime.now().strftime("%H:%M:%S"),
    "CURRENT_DATETIME": lambda: datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
    "CURRENT_YEAR": lambda: str(datetime.datetime.now().year),
    "CURRENT_MONTH": lambda: f"{datetime.datetime.now().month:02}",
    "CURRENT_DAY": lambda: f"{datetime.datetime.now().day:02}",
    "CURRENT_WEEKDAY": lambda: datetime.datetime.now().strftime("%A"),
    "TIMEZONE": lambda: datetime.datetime.now().astimezone().tzname() or "UTC",
    "CURRENT_TIMESTAMP": lambda: str(int(time.time())),
    # platform
    "USERNAME": lambda: os.getlogin(),
    "PLATFORM": lambda: platform.system(),
    "OS_VERSION": lambda: platform.version(),
    "ARCHITECTURE": lambda: platform.machine(),
    "PYTHON_VERSION": lambda: platform.python_version(),
    "HOME": lambda: os.path.expanduser("~"),
    "PROCESS_ID": lambda: str(os.getpid()),
    "WORKING_DIR": lambda: os.getcwd(),
    # network
    "LOCAL_IP": lambda: socket.gethostbyname(socket.gethostname()),
    "PUBLIC_IP": lambda: requests.get("https://api.ipify.org?format=json").json()["ip"],
    "HOSTNAME": lambda: socket.gethostname(),
    # misc
    "RANDOM_INT": lambda: str(random.randint(0, 10000)),
    "RANDOM_HEX": lambda: hex(random.getrandbits(64)),
    "RANDOM_FLOAT": lambda: f"{random.uniform(0, 1):.6f}",
    "RANDOM_STRING": lambda: "".join(random.choices(string.ascii_letters + string.digits, k=10)),
    # clipboard
    "CLIPBOARD": _read_clipboard,
}


def is_builtin_variable(variable_name: str) -> bool:
    global _builtins

    return variable_name in _builtins


def get_builtin_variable_value(variable_name: str) -> str:
    global _builtins

    logger.debug(f"getting builtin variable value for {variable_name}")

    return _builtins[variable_name]()



import pathlib
import unittest
from unittest.mock import MagicMock, patch

import mcp.types as mcp_types

from nerve.models import Configuration
from nerve.runtime import Runtime
from nerve.runtime.runner import Arguments
from nerve.server.mcp import create_mcp_server


class TestMCPServer(unittest.IsolatedAsyncioTestCase):
    @patch("nerve.server.mcp.Runner")
    async def test_mcp_create_server_for_agent_adds_agent_as_tool(self, mock_runner: MagicMock) -> None:
        # Setup
        agent_name = "test_agent"
        config = MagicMock(spec=Configuration)
        config.description = "Test agent description"
        inputs = {"input1": None, "input2": "default_value"}

        run_args = Arguments(
            input_path=pathlib.Path("/path/to/agent"),
            generator="gpt-4",
            conversation_strategy="window",
            conversation_strategy_string="window",
            max_steps=10,
            max_cost=5.0,
            timeout=30,
            quiet=False,
            debug=False,
            log_path=None,
            litellm_debug=False,
            litellm_tracing=None,
            start_state={},
            trace=None,
            task=None,
            interactive=False,
        )

        runtime = MagicMock(spec=Runtime)
        runtime.tools = []

        # Create server
        server = create_mcp_server(
            agent_name=agent_name,
            config=config,
            run_args=run_args,
            inputs=inputs,
            runtime=runtime,
            serve_tools=False,
            tools_only=False,
        )

        list_tools_handler = server.request_handlers[mcp_types.ListToolsRequest]
        server_result = await list_tools_handler(MagicMock())

        self.assertIsInstance(server_result, mcp_types.ServerResult)
        tools_result = server_result.root
        self.assertIsInstance(tools_result, mcp_types.ListToolsResult)

        tools = tools_result.tools  # type: ignore
        self.assertEqual(len(tools), 1)

        tool = tools[0]
        self.assertEqual(tool.name, agent_name)
        self.assertEqual(tool.description, config.description)

        # Verify input schema
        self.assertEqual(tool.inputSchema["type"], "object")
        self.assertEqual(tool.inputSchema["description"], config.description)
        self.assertEqual(tool.inputSchema["required"], ["input1"])  # Only input1 is required
        self.assertIn("input1", tool.inputSchema["properties"])
        self.assertIn("input2", tool.inputSchema["properties"])

    @patch("nerve.server.mcp.Runner")
    async def test_create_mcp_server_for_agent_uses_task_as_description_when_description_not_set(
        self, mock_runner: MagicMock
    ) -> None:
        # Setup
        agent_name = "test_agent"
        config = MagicMock(spec=Configuration)
        config.description = None  # Description not set
        config.task = "Test agent task"  # Task is set
        inputs = {"input1": None, "input2": "default_value"}

        run_args = Arguments(
            input_path=pathlib.Path("/path/to/agent"),
            generator="gpt-4",
            conversation_strategy="window",
            conversation_strategy_string="window",
            max_steps=10,
            max_cost=5.0,
            timeout=30,
            quiet=False,
            debug=False,
            log_path=None,
            litellm_debug=False,
            litellm_tracing=None,
            start_state={},
            trace=None,
            task=None,
            interactive=False,
        )

        runtime = MagicMock(spec=Runtime)
        runtime.tools = []

        # Create server
        server = create_mcp_server(
            agent_name=agent_name,
            config=config,
            run_args=run_args,
            inputs=inputs,
            runtime=runtime,
            serve_tools=False,
            tools_only=False,
        )

        list_tools_handler = server.request_handlers[mcp_types.ListToolsRequest]
        server_result = await list_tools_handler(MagicMock())

        self.assertIsInstance(server_result, mcp_types.ServerResult)
        tools_result = server_result.root
        self.assertIsInstance(tools_result, mcp_types.ListToolsResult)

        tools = tools_result.tools  # type: ignore
        self.assertEqual(len(tools), 1)

        tool = tools[0]
        self.assertEqual(tool.name, agent_name)
        self.assertEqual(tool.description, config.task)  # Description should be the task

        # Verify input schema
        self.assertEqual(tool.inputSchema["type"], "object")
        self.assertEqual(tool.inputSchema["description"], config.task)  # Schema description should also be the task
        self.assertEqual(tool.inputSchema["required"], ["input1"])  # Only input1 is required
        self.assertIn("input1", tool.inputSchema["properties"])
        self.assertIn("input2", tool.inputSchema["properties"])



import pathlib
import typing as t
from enum import Enum

from loguru import logger
from pydantic import AfterValidator, BaseModel, Field
from pydantic_yaml import parse_yaml_raw_as

import nerve


class Mode(str, Enum):
    """
    Running mode.
    """

    INTERACTIVE = "interactive"
    AUTOMATIC = "automatic"


class Status(str, Enum):
    # used in interactive mode
    PAUSED = "paused"
    # the system is running (trying to solve the task)
    RUNNING = "running"
    # the task was completed successfully
    COMPLETED = "completed"
    # the task failed
    FAILED = "failed"

    def is_done(self) -> bool:
        return self in [Status.COMPLETED, Status.FAILED]


class Usage(BaseModel):
    cost: float | None = None
    prompt_tokens: int = 0
    completion_tokens: int = 0
    total_tokens: int = 0

    def __add__(self, other: "Usage") -> "Usage":
        return Usage(
            cost=(self.cost or 0) + (other.cost or 0),
            prompt_tokens=self.prompt_tokens + other.prompt_tokens,
            completion_tokens=self.completion_tokens + other.completion_tokens,
            total_tokens=self.total_tokens + other.total_tokens,
        )


class Tool(BaseModel):
    """
    A tool is a function that can be called by the agent.
    """

    class Argument(BaseModel):
        """
        An argument is a parameter of the tool.
        """

        name: str
        description: str
        example: str = ""

    # if path is set, it'll be loaded from a python file
    path: str | None = None

    # yaml description of the tool
    name: str = ""
    description: str = ""
    arguments: list[Argument] = []
    complete_task: bool = False
    print: bool = False
    mime: str | None = None
    tool: str | None = None


def _check_required_version(required: str | None) -> str | None:
    if required:
        from packaging.requirements import Requirement

        try:
            # a version was specified, convert to a valid requirement
            if required[0].isdigit():
                required_str = f"nerve-adk>={required}"
            else:
                # a full expression was specified, use it as is
                required_str = f"nerve-adk{required}"
            req = Requirement(required_str)
        except Exception as e:
            logger.error(f"error parsing required version '{required}': {e}")
            raise

        if nerve.__version__ not in req.specifier:
            msg = f"required version {required} not satisfied by installed version {nerve.__version__}"
            logger.error(msg)
            raise ValueError(msg)

    return required


class Configuration(BaseModel):
    """
    Configuration for an agent determining its "identity", task and capabilities.
    """

    class MCPServer(BaseModel):
        """
        A MCP server is a server that implements the MCP protocol.
        """

        session_timeout: float = 5

        command: str = "python"
        args: list[str] = []
        env: dict[str, str] = {}
        # for SSE
        url: str | None = None
        headers: dict[str, t.Any] | None = None
        timeout: float = 5
        sse_read_timeout: float = 60 * 5

    class Limits(BaseModel):
        runs: int | None = None
        max_steps: int | None = None
        max_cost: float | None = None
        timeout: int | None = None

    # legacy field used to detect if the user is loading a legacy file
    system_prompt: str | None = Field(default=None, exclude=True)

    # optional generator
    generator: str | None = None
    # optional agent description
    description: str = ""
    # optional nerve version requirement
    requires: t.Annotated[str | None, AfterValidator(_check_required_version)] = None
    # used for versioning the agents
    version: str = "1.0.0"
    # the system prompt, the agent identity
    agent: str | None = None
    # the main agent task
    task: str | None = None
    # optional default values for the initial state
    defaults: dict[str, t.Any] = {}
    # builtin namespaces
    using: list[str] = []
    # jail mechanism for each namespace
    jail: dict[str, list[str]] = {}
    # MCP ( https://modelcontextprotocol.io/ ) servers.
    mcp: dict[str, MCPServer] = {}
    # optional limits
    limits: Limits | None = None
    # custom tooling
    tools: list[Tool | t.Callable[..., t.Any]] = []

    @staticmethod
    def is_agent_config(input_path: pathlib.Path) -> bool:
        try:
            _ = Configuration.from_path(input_path)
            return True
        except Exception:
            return False

    @classmethod
    def from_path(cls, input_path: pathlib.Path) -> "Configuration":
        if input_path.is_dir():
            for option in ("task.yml", "agent.yml"):
                sub_path = input_path / option
                if sub_path.exists():
                    input_path = sub_path
                    break

        with open(input_path) as f:
            return parse_yaml_raw_as(cls, f.read())

    @classmethod
    def from_yml(cls, config_yml: str) -> "Configuration":
        return parse_yaml_raw_as(cls, config_yml)

    @property
    def is_legacy(self) -> bool:
        return self.system_prompt is not None

    def _get_inputs_from_string(self, string: str) -> set[str]:
        from jinja2 import Environment, meta

        return meta.find_undeclared_variables(Environment().parse(string))

    def get_inputs(self) -> dict[str, t.Any]:
        """
        Get the input names for the agent with their default values if set.
        """
        input_names = set()
        tools_names = [t.name if isinstance(t, Tool) else t.__name__ for t in self.tools]

        # from the system prompt
        if self.agent:
            for input_name in self._get_inputs_from_string(self.agent):
                if input_name not in tools_names:
                    input_names.add(input_name)

        # from the task prompt
        if self.task:
            for input_name in self._get_inputs_from_string(self.task):
                if input_name not in tools_names:
                    input_names.add(input_name)

        # from the tools
        for tool in self.tools:
            if isinstance(tool, Tool) and tool.tool:
                arg_names = [arg.name for arg in tool.arguments]
                for input_name in self._get_inputs_from_string(tool.tool):
                    if input_name not in arg_names and input_name not in tools_names:
                        input_names.add(input_name)

        if not self.task:
            input_names.add("task")

        return {input_name: self.defaults.get(input_name) for input_name in input_names}


class Workflow(BaseModel):
    """
    A workflow is a collection of agents that are executed sequentially.
    """

    class Actor(BaseModel):
        """
        An actor is an agent that is part of the workflow.
        """

        generator: str

    name: str
    description: str
    actors: dict[str, Actor]

    @staticmethod
    def is_workflow(input_path: pathlib.Path) -> bool:
        try:
            _ = Workflow.from_path(input_path)
            return True
        except Exception:
            return False

    @classmethod
    def from_path(
        cls,
        input_path: pathlib.Path,
    ) -> "Workflow":
        if input_path.is_dir():
            input_path = input_path / "workflow.yml"

        with open(input_path) as f:
            return parse_yaml_raw_as(cls, f.read())


__all__ = ["Mode", "Status", "Tool", "Configuration", "Workflow"]



import typing as t

import uvicorn
from fastapi import Body, FastAPI, HTTPException, Request
from loguru import logger

from nerve.models import Configuration
from nerve.runtime import Runtime
from nerve.runtime.runner import Arguments, Runner


def _get_input_state_from_request(inputs: dict[str, str], data: dict[str, str]) -> dict[str, str]:
    input_state = inputs.copy()
    for input_name in input_state.keys():
        # get user provided or default value if set
        input_value = data.get(input_name, inputs.get(input_name, None))
        if input_value is None:
            raise HTTPException(status_code=400, detail=f"input '{input_name}' is required")

        input_state[input_name] = input_value

    return input_state


def _create_agent_call_endpoint(
    run_args: Arguments,
    inputs: dict[str, str],
) -> t.Callable[[dict[str, str], Request], t.Coroutine[t.Any, t.Any, t.Any]]:
    logger.debug(f"creating request endpoint for inputs: {inputs}")

    async def _on_request(data: dict[str, str], request: Request) -> t.Any:
        # check if the "raw" query parameter is present
        raw = request.query_params.get("full", "false").lower() == "true"
        client = request.client
        client_host = ""
        if client:
            client_host = client.host

        logger.info(f"request from {client_host}: {data} [raw={raw}]")
        # validate and prepare input state from request
        input_state = _get_input_state_from_request(inputs, data)
        # create a runner
        runner = Runner(run_args, input_state)
        # execute the runner
        output = await runner.run()

        logger.debug(f"output state: {output}")

        if raw:
            return output

        return output.output

    return _on_request


def _create_tool_call_endpoint(
    tool: t.Callable[..., t.Any],
) -> t.Callable[[dict[str, str], Request], t.Coroutine[t.Any, t.Any, dict[str, t.Any]]]:
    async def _on_request(data: dict[str, str] = Body(default=None), request: Request = Request) -> dict[str, t.Any]:  # type: ignore
        client = request.client
        client_host = ""
        if client:
            client_host = client.host

        logger.info(f"request for tool {tool.__name__} from {client_host}: {data}")

        return {"result": await tool(**(data if data else {}))}

    return _on_request


def create_rest_api(
    run_args: Arguments,
    inputs: dict[str, t.Any],
    config: Configuration,
    runtime: Runtime | None,
    serve_tools: bool,
    tools_only: bool,
) -> FastAPI:
    # TODO: use Starlette instead to minimize dependencies
    app = FastAPI()

    if not tools_only:
        logger.info("🌐 creating agent endpoint")
        logger.info("  /")
        app.add_api_route(
            path="/",
            endpoint=_create_agent_call_endpoint(run_args, inputs),
            methods=["POST"],
            response_model=dict,
            summary=config.description,
        )

    if serve_tools and runtime:
        logger.info(f"🌐 creating endpoints for {len(runtime.tools)} tools")
        logger.debug(runtime.tools)

        for tool in runtime.tools:
            logger.info(f"  /{tool.__name__}")
            app.add_api_route(
                path=f"/{tool.__name__}",
                endpoint=_create_tool_call_endpoint(tool),
                methods=["POST"],
                response_model=dict,
                summary=tool.__doc__,
            )

    return app


async def serve_http_app(
    app: t.Any,
    agent_name: str,
    scheme: str,
    host: str,
    port: int,
    debug: bool,
) -> None:
    logger.info(f"🌐 serving {agent_name} on {scheme}://{host}:{port}/ ...")

    config = uvicorn.Config(app, host=host, port=port, log_level="debug" if debug else "warning")
    server = uvicorn.Server(config)

    await server.serve()



import asyncio

import typer
from loguru import logger
from typer_di import Depends, TyperDI

import nerve
from nerve.cli.utils import _get_run_args
from nerve.models import Configuration, Mode, Workflow
from nerve.runtime import logging, state
from nerve.runtime.agent import Agent
from nerve.runtime.flow import Flow
from nerve.runtime.runner import Arguments

cli = TyperDI(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": ["-h", "--help"]},
    help="Execute an agent or a workflow.",
)
def run(
    ctx: typer.Context,
    args: Arguments = Depends(_get_run_args),
) -> None:
    logging.init(
        args.log_path,
        level="DEBUG" if args.debug else "SUCCESS" if args.quiet else "INFO",
        litellm_debug=args.litellm_debug,
        litellm_tracing=args.litellm_tracing,
    )
    logger.info(f"🧠 nerve v{nerve.__version__}")

    asyncio.run(
        _run(
            ctx.args,
            args,
        )
    )


def _get_start_state_from_args(args: list[str]) -> dict[str, str]:
    # any unknown argument will populate the start_state
    start_state = {}
    for i in range(0, len(args), 2):
        key = args[i].removeprefix("--").removeprefix("-").replace("-", "_")
        value = args[i + 1] if i + 1 < len(args) else ""
        start_state[key] = value
    return start_state


async def _run(extra_args: list[str], args: Arguments) -> None:
    if args.trace:
        state.set_trace_file(args.trace)

    if args.interactive:
        state.set_mode(Mode.INTERACTIVE)

    # make variables available to the runtime
    start_state = args.start_state
    start_state.update(_get_start_state_from_args(extra_args))
    state.update_variables(start_state)

    # check if input_path is a workflow or single agent
    if Workflow.is_workflow(args.input_path):
        # full workflow
        flow = await Flow.from_path(
            args.input_path,
            window_strategy=args.conversation_strategy,
            max_steps=args.max_steps,
            max_cost=args.max_cost,
            timeout=args.timeout,
            start_state=start_state,
        )

    elif Configuration.is_agent_config(args.input_path):
        # single agent
        flow = await Flow.build(
            actors=[await Agent.create_from_file(args.generator, args.input_path, args.conversation_strategy)],
            max_steps=args.max_steps,
            max_cost=args.max_cost,
            timeout=args.timeout,
            start_state=start_state,
        )

    else:
        logger.error(f"path '{args.input_path}' is not a valid workflow or agent configuration")
        raise typer.Abort()

    await flow.run(args.task)

    logger.debug("exiting")



import pathlib
import shutil
import typing as t

import typer

import nerve
from nerve.defaults import (
    DEFAULT_AGENTS_LOAD_PATH,
)

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help=f"Uninstall an agent or workflow from {DEFAULT_AGENTS_LOAD_PATH}",
)
def uninstall(
    name: t.Annotated[
        str,
        typer.Argument(help="Name of the agent or workflow to uninstall."),
    ],
    yes: t.Annotated[
        bool,
        typer.Option("--yes", "-y", help="Uninstall without asking for confirmation."),
    ] = False,
) -> None:
    print(f"🧠 nerve v{nerve.__version__}")

    # make sure no funny business is happening
    if ".." in name:
        print(f"🚨 {name} is not a valid agent or workflow name")
        exit(1)

    # get the path to the agent or workflow
    path = pathlib.Path(DEFAULT_AGENTS_LOAD_PATH) / name
    if not path.exists():
        print(f"❌ {name} does not exist in {DEFAULT_AGENTS_LOAD_PATH}")
        exit(1)

    # ask for confirmation
    if not yes:
        typer.confirm(
            f"⚠️  Are you sure you want to delete {name} from {DEFAULT_AGENTS_LOAD_PATH}? This action is irreversible.",
            abort=True,
        )

    # uninstall the agent or workflow
    shutil.rmtree(path)
    print(f"🧠 {name} uninstalled from {DEFAULT_AGENTS_LOAD_PATH}")



import unittest

from nerve.runtime.runner import _parse_events


class TestParseEvents(unittest.TestCase):
    def test_task_completed(self) -> None:
        """Test when task is completed successfully."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {"name": "task_complete", "timestamp": 1.0, "data": {"output": "result"}},
            {"name": "flow_complete", "timestamp": 2.0, "data": {"steps": 3, "usage": {"tokens": 100}}},
        ]

        result = _parse_events(inputs, events)

        self.assertTrue(result.task_success)
        self.assertEqual(result.output_object, {"output": "result"})
        self.assertEqual(result.steps, 3)
        self.assertEqual(result.time, 2.0)
        self.assertEqual(result.usage, {"tokens": 100})

    def test_task_completed_with_reason(self) -> None:
        """Test when task is completed with a reason."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {"name": "task_complete", "timestamp": 1.0, "data": {"reason": "Task finished successfully"}},
            {"name": "flow_complete", "timestamp": 2.0, "data": {"steps": 2, "usage": {"tokens": 50}}},
        ]

        result = _parse_events(inputs, events)

        self.assertTrue(result.task_success)
        self.assertEqual(result.output_object, {"reason": "Task finished successfully"})
        self.assertEqual(result.steps, 2)
        self.assertEqual(result.time, 2.0)
        self.assertEqual(result.usage, {"tokens": 50})

    def test_task_failed(self) -> None:
        """Test when task fails."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {"name": "task_failed", "timestamp": 1.0, "data": {"error": "Something went wrong"}},
            {"name": "flow_complete", "timestamp": 2.0, "data": {"steps": 1, "usage": {"tokens": 20}}},
        ]

        result = _parse_events(inputs, events)

        self.assertFalse(result.task_success)
        self.assertEqual(result.output_object, {"error": "Something went wrong"})
        self.assertEqual(result.steps, 1)
        self.assertEqual(result.time, 2.0)
        self.assertEqual(result.usage, {"tokens": 20})

    def test_task_failed_with_reason(self) -> None:
        """Test when task fails with a reason."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {"name": "task_failed", "timestamp": 1.0, "data": {"reason": "Invalid input"}},
            {"name": "flow_complete", "timestamp": 2.0, "data": {"steps": 1, "usage": {"tokens": 20}}},
        ]

        result = _parse_events(inputs, events)

        self.assertFalse(result.task_success)
        self.assertEqual(result.output_object, {"reason": "Invalid input"})
        self.assertEqual(result.steps, 1)
        self.assertEqual(result.time, 2.0)
        self.assertEqual(result.usage, {"tokens": 20})

    def test_flow_completed_with_variables(self) -> None:
        """Test when flow completes with variables in state."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {
                "name": "flow_complete",
                "timestamp": 2.0,
                "data": {
                    "steps": 4,
                    "usage": {"tokens": 150},
                    "state": {"variables": {"input1": "value1", "output1": "result1"}},
                },
            },
        ]

        result = _parse_events(inputs, events)

        self.assertEqual(result.output_object, {"output1": "result1"})
        self.assertEqual(result.steps, 4)
        self.assertEqual(result.time, 2.0)
        self.assertEqual(result.usage, {"tokens": 150})

    def test_fallback_to_text_response(self) -> None:
        """Test fallback to text response when no task completion events."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {"name": "text_response", "timestamp": 1.0, "data": {"response": "This is a text response"}},
        ]

        result = _parse_events(inputs, events)

        self.assertEqual(result.output_object, {"response": "This is a text response"})
        self.assertEqual(result.steps, 0)
        self.assertEqual(result.time, 0.0)
        self.assertEqual(result.usage, {})

    def test_fallback_to_tool_called(self) -> None:
        """Test fallback to tool called when no task completion or text response."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
            {"name": "tool_called", "timestamp": 1.0, "data": {"result": "Tool execution result"}},
        ]

        result = _parse_events(inputs, events)

        self.assertEqual(result.output_object, {"output": "Tool execution result"})
        self.assertEqual(result.steps, 0)
        self.assertEqual(result.time, 0)
        self.assertEqual(result.usage, {})

    def test_no_relevant_events(self) -> None:
        """Test when no relevant events are found."""
        inputs = {"input1": "value1"}
        events = [
            {"name": "flow_start", "timestamp": 0.0},
        ]

        result = _parse_events(inputs, events)

        self.assertIsNone(result.output_object)
        self.assertFalse(result.task_success)
        self.assertEqual(result.steps, 0)
        self.assertEqual(result.time, 0.0)
        self.assertEqual(result.usage, {})



import asyncio
import json
import typing as t

import mcp.types as mcp_types
from fastapi import HTTPException
from loguru import logger
from mcp import stdio_server
from mcp.server import Server
from mcp.server.sse import SseServerTransport
from starlette.applications import Starlette
from starlette.routing import Mount, Route

from nerve.models import Configuration
from nerve.runtime import Runtime
from nerve.runtime.runner import Arguments, Output, Runner
from nerve.tools.protocol import get_tool_schema


def _get_input_state_from_request(inputs: dict[str, str | None], data: dict[str, str]) -> dict[str, str]:
    input_state: dict[str, str] = {}
    for input_name, default_value in inputs.items():
        # get user provided or default value if set
        input_value = data.get(input_name, default_value)
        if input_value is None:
            raise HTTPException(status_code=400, detail=f"input '{input_name}' is required")

        input_state[input_name] = input_value

    return input_state


def _create_listing_handler(
    tools: list[mcp_types.Tool],
) -> t.Callable[[t.Any], t.Coroutine[t.Any, t.Any, mcp_types.ServerResult]]:
    async def _handler(_: t.Any) -> mcp_types.ServerResult:
        return mcp_types.ServerResult(mcp_types.ListToolsResult(tools=tools))

    return _handler


async def _handle_agent_call(
    run_args: Arguments,
    inputs: dict[str, str | None],
    agent_name: str,
    mcp_server: Server,  # type: ignore
    arguments: dict[str, t.Any],
) -> Output:
    input_state = _get_input_state_from_request(inputs, arguments)
    runner = Runner(run_args, input_state)
    runner.set_stdout_fn(
        lambda x: asyncio.create_task(
            mcp_server.request_context.session.send_log_message(
                level="info",
                data=x,
                logger=agent_name,
            )
        )
    )
    runner.set_stderr_fn(
        lambda x: asyncio.create_task(
            mcp_server.request_context.session.send_log_message(
                level="error",
                data=x,
                logger=agent_name,
            )
        )
    )
    # execute the runner
    return await runner.run()


def _mcp_text_result(data: str, error: bool = False) -> mcp_types.ServerResult:
    return mcp_types.ServerResult(
        mcp_types.CallToolResult(content=[mcp_types.TextContent(type="text", text=data)], isError=error)
    )


def _mcp_error_result(error: Exception) -> mcp_types.ServerResult:
    return mcp_types.ServerResult(
        mcp_types.CallToolResult(content=[mcp_types.TextContent(type="text", text=str(error))], isError=True)
    )


def _create_call_handler(
    run_args: Arguments,
    inputs: dict[str, str | None],
    agent_name: str,
    mcp_server: Server,  # type: ignore
    runtime: Runtime | None,
) -> t.Callable[[mcp_types.CallToolRequest], t.Coroutine[t.Any, t.Any, mcp_types.ServerResult]]:
    tools_dict = {tool.__name__: tool for tool in runtime.tools} if runtime else {}

    async def _handler(req: mcp_types.CallToolRequest) -> mcp_types.ServerResult:
        nonlocal tools_dict

        try:
            arguments = req.params.arguments or {}
            tool = tools_dict.get(req.params.name, None)
            if tool:
                # execute a tool
                return _mcp_text_result(json.dumps(await tool(**arguments)))
            elif req.params.name == agent_name:
                # execute the agent itself
                call_result = await _handle_agent_call(
                    run_args,
                    inputs,
                    agent_name,
                    mcp_server,
                    arguments,
                )
                return _mcp_text_result(json.dumps(call_result.output))
            else:
                raise Exception(f"unknown tool: {req.params.name}")

        except Exception as e:
            return _mcp_error_result(e)

    return _handler


def create_mcp_server(
    agent_name: str,
    config: Configuration,
    run_args: Arguments,
    inputs: dict[str, str | None],
    runtime: Runtime | None,
    serve_tools: bool = False,
    tools_only: bool = False,
) -> Server:  # type: ignore
    server = Server(agent_name)  # type: ignore
    tools: list[mcp_types.Tool] = []

    # create a tool for the agent itself
    if not tools_only:
        logger.info(f"🧰 creating MCP tool for {agent_name}")
        agent_description = config.description or config.task
        tools.append(
            mcp_types.Tool(
                name=agent_name,
                description=agent_description,
                inputSchema={
                    "type": "object",
                    "required": [name for name, default_value in inputs.items() if default_value is None],
                    "description": agent_description,
                    "properties": {
                        name: {
                            "type": "string",
                            "description": "",  # TODO: anything we can add here?
                        }
                        for name in inputs.keys()
                    },
                },
            )
        )

    # create a tool for each tool in the runtime
    if serve_tools and runtime:
        logger.info(f"🧰 creating MCP tools for {len(runtime.tools)} agent tools")
        for tool in runtime.tools:
            input_schema = get_tool_schema(run_args.generator, tool).get("function", {}).get("parameters", {})
            logger.info(f"  {tool.__name__}")
            tools.append(
                mcp_types.Tool(
                    name=tool.__name__,
                    description=tool.__doc__,
                    inputSchema=input_schema,
                )
            )

    server.request_handlers[mcp_types.ListToolsRequest] = _create_listing_handler(tools)
    server.request_handlers[mcp_types.CallToolRequest] = _create_call_handler(
        run_args,
        inputs,
        agent_name,
        server,
        runtime,
    )

    return server


def create_sse_app(debug: bool, server: Server) -> Starlette:  # type: ignore
    sse = SseServerTransport("/messages/")

    async def handle_sse(request: t.Any) -> t.Any:
        async with sse.connect_sse(request.scope, request.receive, request._send) as streams:
            await server.run(streams[0], streams[1], server.create_initialization_options())

    return Starlette(
        debug=debug,
        routes=[
            Route("/sse", endpoint=handle_sse),
            Mount("/messages/", app=sse.handle_post_message),
        ],
    )


async def serve_stdio_app(server: Server, agent_name: str) -> None:  # type: ignore
    async with stdio_server() as streams:
        logger.info(f"🌐 serving {agent_name} on stdout ...")
        await server.run(streams[0], streams[1], server.create_initialization_options())



import asyncio
import atexit
import json
import os
import pathlib
import sys
import time
import typing as t
import uuid

from loguru import logger
from pydantic import BaseModel

from nerve.defaults import DEFAULT_RUNS_PATH


class Arguments(BaseModel):
    input_path: pathlib.Path
    task: str | None
    generator: str
    conversation_strategy_string: str
    conversation_strategy: t.Any
    interactive: bool
    debug: bool
    litellm_debug: bool
    litellm_tracing: str | None
    quiet: bool
    max_steps: int
    max_cost: float
    timeout: int | None
    log_path: pathlib.Path | None
    trace: pathlib.Path | None
    start_state: dict[str, t.Any]

    def to_serializable(self) -> dict[str, t.Any]:
        return {
            k: v if not isinstance(v, pathlib.Path) else v.as_posix()
            for k, v in self.model_dump().items()
            if k != "conversation_strategy"
        }


def _create_command_line(
    run_args: Arguments,
    input_state: dict[str, str],
    events_file: pathlib.Path,
) -> list[str]:
    nerve_bin = sys.argv[0]
    command_line = [
        nerve_bin,
        "run",
        str(run_args.input_path),
        "--generator",
        run_args.generator,
        "--conversation",
        run_args.conversation_strategy_string,
        "--max-steps",
        str(run_args.max_steps),
        "--max-cost",
        str(run_args.max_cost),
    ]

    if run_args.timeout:
        command_line.append("--timeout")
        command_line.append(str(run_args.timeout))

    # if run_args.quiet:
    #     command_line.append("--quiet")

    # if run_args.debug:
    #     command_line.append("--debug")

    # if run_args.litellm_debug:
    #     command_line.append("--litellm-debug")

    if run_args.litellm_tracing:
        command_line.append("--litellm-tracing")
        command_line.append(run_args.litellm_tracing)

    # if the task is set, add it to the command line
    if "task" in input_state:
        command_line.append("--task")
        command_line.append(input_state["task"])
        del input_state["task"]

    command_line.append("--start-state")
    command_line.append(json.dumps(input_state))

    command_line.append("--trace")
    command_line.append(str(events_file))

    logger.debug(f"command_line: {command_line}")
    return command_line


def _get_last_event_with_name(events: list[dict[str, t.Any]], name: str) -> dict[str, t.Any] | None:
    for event in reversed(events):
        if event["name"] == name:
            return event
    return None


class ParsedEvents(BaseModel):
    output_object: dict[str, t.Any] | None
    task_success: bool
    steps: int
    time: float
    usage: dict[str, t.Any]


def _parse_events(inputs: dict[str, str], events: list[dict[str, t.Any]]) -> ParsedEvents:
    started_at = events[0].get("timestamp", 0.0)
    # we'll need this either case
    flow_completed = _get_last_event_with_name(events, "flow_complete")
    # prepare the output object
    parsed = ParsedEvents(output_object=None, task_success=False, steps=0, time=0, usage={})
    # set what we can for now
    if flow_completed is not None:
        parsed.steps = flow_completed.get("data", {}).get("steps", 0)
        parsed.usage = flow_completed.get("data", {}).get("usage", {})
        parsed.time = flow_completed.get("timestamp", 0.0) - started_at
        logger.debug(f"flow_completed: {parsed.steps} steps, {parsed.time}s, {parsed.usage}")

    # one of the tools wrote an output variable and set the task to complete
    task_completed = _get_last_event_with_name(events, "task_complete")
    if task_completed is not None:
        data = task_completed.get("data", {})
        reason = data.get("reason", {})
        if reason:
            parsed.output_object = {"reason": reason}
        else:
            parsed.output_object = data

        parsed.task_success = True
        return parsed

    # task failed
    task_failed = _get_last_event_with_name(events, "task_failed")
    if task_failed is not None:
        data = task_failed.get("data", {})
        reason = data.get("reason", {})
        if reason:
            parsed.output_object = {"reason": reason}
        else:
            parsed.output_object = data

        parsed.task_success = False
        return parsed

    # the flow completed successfully and a variable has been written (by the tool
    # that completed the task) to the output state. this is in theory redundant, but
    # we keep it for now to be safe
    if flow_completed:
        variables = flow_completed.get("data", {}).get("state", {}).get("variables", {})
        parsed.output_object = {name: value for name, value in variables.items() if name not in inputs}
        return parsed

    # fallback to the latest tool call output or text response
    # whatever comes first
    for event in reversed(events):
        if event["name"] == "text_response":
            parsed.output_object = {"response": event["data"]["response"]}
            return parsed

        elif event["name"] == "tool_called":
            parsed.output_object = {"output": event["data"]["result"]}
            return parsed

    return parsed


async def _default_stdout_fn(x: str) -> None:
    logger.debug(x)


async def _default_stderr_fn(x: str) -> None:
    logger.debug(x)


class Output(BaseModel):
    generated_at: float
    command_line: list[str]
    exit_code: int
    stdout: list[str]
    stderr: list[str]
    events: list[dict[str, t.Any]]
    output: dict[str, t.Any]
    task_success: bool
    steps: int
    time: float
    usage: dict[str, t.Any]


class Runner:
    def __init__(
        self,
        args: Arguments,
        input_state: dict[str, str] | None = None,
        id: str | None = None,
        base_path: pathlib.Path = DEFAULT_RUNS_PATH,
        clean_at_exit: bool = True,
    ):
        if not base_path.exists():
            base_path.mkdir(parents=True, exist_ok=True)

        self.id = id or str(uuid.uuid4())
        self.events_file = base_path / f"run-{self.id}.jsonl"
        self.input_state = input_state or {}
        self.command_line = _create_command_line(
            args,
            self.input_state,
            self.events_file,
        )
        self._stdout_fn: t.Callable[[str], t.Awaitable[None]] = _default_stdout_fn
        self._stderr_fn: t.Callable[[str], t.Awaitable[None]] = _default_stderr_fn
        self._process: asyncio.subprocess.Process | None = None

        if clean_at_exit:
            atexit.register(self._clean_up)

    def _clean_up(self) -> None:
        if self.events_file.exists():
            logger.debug(f"removing events file {self.events_file}")
            if self._process is not None:
                self._process.kill()
                self._process = None
            self.events_file.unlink()

    def set_stdout_fn(self, fn: t.Callable[[str], t.Awaitable[None]]) -> None:
        self._stdout_fn = fn

    def set_stderr_fn(self, fn: t.Callable[[str], t.Awaitable[None]]) -> None:
        self._stderr_fn = fn

    async def run(self) -> Output:
        logger.debug(f"spawning runner {self.id} for inputs: {self.input_state}")

        generated_at = time.time()
        outerr: dict[str, list[str]] = {
            "stdout": [],
            "stderr": [],
        }

        async def read_stream(stream: asyncio.StreamReader | None, name: str) -> None:
            nonlocal outerr

            while True and stream is not None:
                line = await stream.readline()
                if not line:
                    break

                if name == "stdout":
                    await self._stdout_fn(line.decode().rstrip())
                else:
                    await self._stderr_fn(line.decode().rstrip())

                outerr[name].append(line.decode().rstrip())

        self._process = await asyncio.create_subprocess_exec(
            *self.command_line,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            env=os.environ.copy(),
        )
        stdout_task = asyncio.create_task(read_stream(self._process.stdout, "stdout"))
        stderr_task = asyncio.create_task(read_stream(self._process.stderr, "stderr"))

        # wait for the process and stdout/stderr readers to complete
        await self._process.wait()
        await asyncio.gather(stdout_task, stderr_task)

        logger.debug(f"process exited with code {self._process.returncode}, reading events ...")

        # read the events file
        events = []
        with open(self.events_file) as f:
            for line in f:
                events.append(json.loads(line))

        logger.debug(f"read {len(events)} events")

        events.sort(key=lambda event: event.get("timestamp", 0), reverse=False)

        parsed = _parse_events(self.input_state, events)
        if parsed.output_object is None:
            logger.warning(f"could not get raw output value from runner {self.id}")

            if outerr["stderr"]:
                parsed.output_object = {"output": "\n".join(outerr["stderr"])}
            elif outerr["stdout"]:
                parsed.output_object = {"output": "\n".join(outerr["stdout"])}
            else:
                parsed.output_object = {"output": "the tool did not write any output"}

        logger.debug(f"output value: {parsed.output_object}")

        exit_code = self._process.returncode or 0
        self._process = None

        return Output(
            generated_at=generated_at,
            command_line=self.command_line,
            exit_code=exit_code,
            stdout=outerr["stdout"],
            stderr=outerr["stderr"],
            events=events,
            output=parsed.output_object,
            task_success=parsed.task_success,
            steps=parsed.steps,
            time=parsed.time,
            usage=parsed.usage,
        )



import asyncio
import pathlib
import sys
import typing as t

import typer
from loguru import logger
from typer_di import Depends, TyperDI

import nerve
from nerve.cli.utils import _get_run_args
from nerve.defaults import (
    DEFAULT_SERVE_HOST,
    DEFAULT_SERVE_PORT,
)
from nerve.models import Configuration
from nerve.runtime import Runtime, logging
from nerve.runtime.runner import Arguments
from nerve.server.mcp import create_mcp_server, create_sse_app, serve_stdio_app
from nerve.server.rest import create_rest_api, serve_http_app

cli = TyperDI(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


def _get_agent_name(input_path: pathlib.Path) -> str:
    stem = input_path.stem
    if stem in ("agent", "workflow", "task"):
        return input_path.parent.name
    return stem


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    no_args_is_help=True,
    help="Serve an agent as a REST API or MCP server.",
)
def serve(
    run_args: Arguments = Depends(_get_run_args),
    host: t.Annotated[
        str,
        typer.Option("--host", "-h", help="Bind host to serve the agent on."),
    ] = DEFAULT_SERVE_HOST,
    port: t.Annotated[
        int,
        typer.Option("--port", "-p", help="Bind port to serve the agent on."),
    ] = DEFAULT_SERVE_PORT,
    mcp: t.Annotated[
        bool,
        typer.Option("--mcp", help="Start as MCP server."),
    ] = False,
    mcp_sse: t.Annotated[
        bool,
        typer.Option("--sse", help="Start as MCP server with SSE transport."),
    ] = False,
    serve_tools: t.Annotated[
        bool,
        typer.Option(
            "--tools",
            "-t",
            help="Serve tools as MCP servers. Automatically enabled if agent doesn't have a system prompt.",
        ),
    ] = False,
    tools_only: t.Annotated[
        bool,
        typer.Option("--tools-only", help="Serve tools only."),
    ] = False,
) -> None:
    # log to stderr instead of stdout if we're running as MCP server without SSE
    log_target = sys.stderr if mcp and not mcp_sse else sys.stdout
    logging.init(
        run_args.log_path,
        level="DEBUG" if run_args.debug else "INFO",
        litellm_debug=run_args.litellm_debug,
        target=log_target,
    )
    logger.info(f"🧠 nerve v{nerve.__version__}")

    asyncio.run(_serve(host, port, mcp, mcp_sse, serve_tools, tools_only, run_args))


async def _serve(
    host: str,
    port: int,
    mcp: bool,
    mcp_sse: bool,
    serve_tools: bool,
    tools_only: bool,
    run_args: Arguments,
) -> None:
    # validate and collect inputs from the agent
    if not Configuration.is_agent_config(run_args.input_path):
        logger.error(f"path '{run_args.input_path}' is not a valid agent configuration")
        raise typer.Abort()

    logger.debug(f"loading agent from {run_args.input_path}")
    config = Configuration.from_path(run_args.input_path)
    agent_name = _get_agent_name(run_args.input_path)
    logger.debug(f"agent {agent_name} loaded: {config.description}")

    inputs = config.get_inputs()
    logger.debug(f"creating endpoint for inputs: {inputs}")
    runtime: Runtime | None = None

    if tools_only or not config.system_prompt and not config.agent and not config.task:
        logger.info("🧰 tools-only mode")
        serve_tools = True
        tools_only = True
    elif serve_tools:
        logger.info("🧠 + 🧰 serving agent and tools")
    else:
        logger.info("🧠 serving agent")

    if serve_tools:
        # if we have to serve tools, we need to build the runtime
        runtime = await Runtime.build(
            working_dir=run_args.input_path if run_args.input_path.is_dir() else run_args.input_path.parent,
            name=agent_name,
            configuration=config,
            debug=run_args.debug,
        )

    if mcp or mcp_sse:
        # MCP server
        server = create_mcp_server(
            agent_name,
            config,
            run_args,
            inputs,
            runtime,
            serve_tools,
            tools_only,
        )

        if mcp_sse:
            # via SSE (http)
            app = create_sse_app(run_args.debug, server)

            await serve_http_app(app, agent_name, "sse", host, port, run_args.debug)

        else:
            # via stdout (as a process)
            await serve_stdio_app(server, agent_name)

    else:
        # start as REST API
        app = create_rest_api(
            run_args,
            inputs,
            config,
            runtime,
            serve_tools,
            tools_only,
        )

        await serve_http_app(app, agent_name, "http", host, port, run_args.debug)



import os
import pathlib
import shutil
import tempfile
import typing as t
import zipfile

import requests
import typer

import nerve
from nerve.defaults import (
    DEFAULT_AGENTS_LOAD_PATH,
)
from nerve.models import Configuration, Workflow

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


def _download_and_unzip(url: str) -> pathlib.Path:
    temp_dir = pathlib.Path(tempfile.mkdtemp())
    local_zip_path = pathlib.Path(os.path.join(temp_dir, "archive.zip"))

    print(f"🌐 downloading {url} ...")

    # download to temporary file
    with requests.get(url, stream=True, verify=True, allow_redirects=True) as response:
        response.raise_for_status()
        with open(local_zip_path, "wb") as zip_file:
            for chunk in response.iter_content(chunk_size=8192):
                zip_file.write(chunk)

    # unzip to temporary directory
    try:
        with zipfile.ZipFile(local_zip_path, "r") as zf:
            for member in zf.infolist():
                file_path = os.path.realpath(os.path.join(temp_dir, member.filename))
                if file_path.startswith(os.path.realpath(temp_dir)):
                    zf.extract(member, temp_dir)
                else:
                    raise Exception("Attempted Path Traversal Attack Detected")

    finally:
        # always remove the zip file
        if local_zip_path.exists():
            os.remove(local_zip_path)

    # in the case of a github repo we expect a single folder named username-repo-commit
    elems = list(temp_dir.iterdir())
    if len(elems) == 1 and elems[0].is_dir():
        temp_dir = elems[0]

    return temp_dir


def _get_source_path_type(source_path: pathlib.Path) -> str | None:
    if Configuration.is_agent_config(source_path):
        return "agent"
    elif Workflow.is_workflow(source_path):
        return "workflow"
    else:
        return None


def _install_from_path(source_path: pathlib.Path, target_path: pathlib.Path, overwrite: bool = False) -> None:
    if not source_path.exists():
        print(f"❌ {source_path} does not exist")
        exit(1)

    if not source_path.is_dir():
        print(f"❌ {source_path} is not a directory")
        exit(1)

    source_path_type = _get_source_path_type(source_path)
    if source_path_type is None:
        print(f"❌ {source_path} is not a valid agent or workflow")
        exit(1)

    target_agent_path = target_path / source_path.name
    if target_agent_path.exists():
        if not overwrite:
            if input(f"⚠️  {target_agent_path} already exists, do you want to overwrite it? [y/N] ") != "y":
                print("❌ installation aborted")
                exit(1)

        shutil.rmtree(target_agent_path)

    target_path.mkdir(parents=True, exist_ok=True)

    shutil.copytree(source_path, target_agent_path)

    print(f"✅ {source_path_type} installed to {target_agent_path}, use `nerve run {source_path.name}` to execute it")


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="Install an agent or workflow from a git repository or local directory.",
)
def install(
    source: t.Annotated[
        str,
        typer.Argument(
            help="Source to install from. It can be a git repository (full url or username/repo) or a local directory."
        ),
    ],
    path: t.Annotated[
        pathlib.Path,
        typer.Option("--path", "-p", help="Path to install to."),
    ] = DEFAULT_AGENTS_LOAD_PATH,
    overwrite: t.Annotated[
        bool,
        typer.Option("--overwrite", "-o", help="Overwrite existing agent or workflow without asking for confirmation."),
    ] = False,
    branch: t.Annotated[
        str,
        typer.Option("--branch", "-b", help="Branch to install from if the source is a git repository."),
    ] = "main",
) -> None:
    print(f"🧠 nerve v{nerve.__version__}")

    source_path_is_tmp = False
    source_path = pathlib.Path(source)

    if not source_path.exists():
        zip_file_url: str = ""
        final_name: str = ""

        if source.startswith("https://github.com"):
            # https://github.com/username/repo (github repository full url)
            zip_file_url = f"{source}/zipball/{branch}"
            final_name = source.split("/")[-1]
        elif source.startswith("https://") and source.endswith(".zip"):
            # https://whatever.dev/agent.zip (zip file url)
            zip_file_url = source
            final_name = source.split("/")[-1].replace(".zip", "")
        elif source.count("/") == 1:
            # username/repo (github repository)
            zip_file_url = f"https://github.com/{source}/zipball/{branch}"
            final_name = source.split("/")[-1]
        else:
            print(f"❌ {source} is not a valid source")
            exit(1)

        try:
            # download, unzip and rename to final name if needed
            temp_dir = _download_and_unzip(zip_file_url)
            if temp_dir.name != final_name:
                new_temp_dir = temp_dir.parent / final_name
                shutil.move(temp_dir, new_temp_dir)
                temp_dir = new_temp_dir
        except Exception as e:
            print(f"❌ error: {e}")
            exit(1)

        source_path_is_tmp = True
        source_path = temp_dir

    try:
        _install_from_path(source_path, path, overwrite)
    except Exception as e:
        print(f"❌ error: {e}")
        exit(1)
    finally:
        if source_path_is_tmp:
            shutil.rmtree(source_path)



import pathlib
import typing as t

import natsort
import requests
import typer
from termcolor import colored

import nerve
from nerve.defaults import (
    DEFAULT_AGENTS_LOAD_PATH,
)
from nerve.models import Configuration, Workflow

cli = typer.Typer(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


@cli.command(
    context_settings={"help_option_names": ["-h", "--help"]},
    help="List locally installed agents and available ones from the awesomeagents.ai index.",
)
def agents(
    path: t.Annotated[
        pathlib.Path,
        typer.Argument(help="Path to the agent or workflow to create"),
    ] = DEFAULT_AGENTS_LOAD_PATH,
    offline: t.Annotated[
        bool,
        typer.Option("--offline", "-o", help="Only show installed agents."),
    ] = False,
) -> None:
    print(f"🧠 nerve v{nerve.__version__}")

    _show_installed_agents(path)

    if not offline:
        agents = _fetch_awesome_agents()
        if agents:
            _show_awesome_agents(agents)


def _fetch_awesome_agents() -> list[dict[str, t.Any]] | None:
    try:
        response = requests.get("https://api.awesomeagents.ai/index.json")
        response.raise_for_status()
        agents = response.json()
        return [agent for agent in agents if "nerve" in agent.get("stack", [])]
    except Exception:
        return None


def _show_awesome_agents(agents: list[dict[str, t.Any]]) -> None:
    print("🔍 Available from the index:\n")
    for agent in agents:
        repo = agent["repo"]
        # Extract username/repo from the repository URL
        repo_parts = repo.split("/")
        if len(repo_parts) >= 2:
            username_repo = f"{repo_parts[-2]}/{repo_parts[-1]}"
        else:
            username_repo = repo
        print(f"     {colored(username_repo, 'white', attrs=['bold'])} - {agent['description']}")

    print()


def _show_installed_agents(path: pathlib.Path) -> None:
    anything = False

    if path.exists() and path.is_dir():
        print()
        print(f"📁 Installed in {path.absolute()}:\n")

        items = []
        for item in path.iterdir():
            items.append(item)

        for item in natsort.natsorted(items):
            if Workflow.is_workflow(item):
                workflow = Workflow.from_path(item)
                print(
                    f"     {colored(item.name, 'white', attrs=['bold'])} "
                    + colored("<workflow>", "blue")
                    + f" - {workflow.description}"
                )
                anything = True
            elif Configuration.is_agent_config(item):
                config = Configuration.from_path(item)
                print(
                    f"     {colored(item.name, 'white', attrs=['bold'])} "
                    + colored("<agent>", "green")
                    + f" - {config.description}"
                )
                anything = True

        print()

    if not anything:
        print(colored(f"No agents or workflows found in {path}", "light_grey"))



import typing as t

from loguru import logger

from nerve.models import Usage


# until this is not fixed, ollama needs special treatment: https://github.com/BerriAI/litellm/issues/6353
class OllamaGlue:
    def __init__(self, api_base: str, generator_id: str, generator_params: dict[str, t.Any]) -> None:
        import ollama

        self.model = "/".join(generator_id.split("/")[1:])
        self.client = ollama.AsyncClient(host=api_base)
        self.generator_params = generator_params

        logger.debug(f"using ollama client for model {self.model}")

    async def _process_conversation(self, conversation: list[dict[str, t.Any]]) -> list[dict[str, t.Any]]:
        fixed = []

        """
        Ollama needs this:

        {
            'role': 'user',
            'content': [
                {'type': 'text', 'text': 'read_webcam_image returned the following response:'},
                {'type': 'image_url', 'image_url': {'url': 'data:image/jpeg;base64,/9j/4AAQSkZJRgABAQAAAQABAAD/2wBDAAIBAQEBA....'}}
            ]
        }

        To be fixed into this:

        {
            'role': 'user',
            'content': 'read_webcam_image returned the following response:',
            'images': [ raw image data ]

        }
        """

        for message in conversation:
            if "content" in message and isinstance(message["content"], list):
                if (
                    len(message["content"]) == 2
                    and message["content"][0]["type"] == "text"
                    and message["content"][1]["type"] == "image_url"
                ):
                    fixed.append(
                        {
                            "role": "user",
                            "content": message["content"][0]["text"],
                            "images": [message["content"][1]["image_url"]["url"].split(",")[1]],
                        }
                    )
            else:
                fixed.append(message)

        return fixed

    async def generate(
        self, conversation: list[dict[str, t.Any]], tools_schema: list[dict[str, t.Any]] | None
    ) -> tuple[Usage, t.Any]:
        conversation = await self._process_conversation(conversation)

        logger.debug(f"ollama.conversation: {conversation}")
        response = await self.client.chat(
            model=self.model,
            messages=conversation,
            tools=tools_schema,
            **self.generator_params,
        )
        logger.debug(f"ollama.response: {response}")
        return Usage(
            prompt_tokens=0,
            completion_tokens=0,
            total_tokens=0,
        ), response.message



# ruff: noqa: B010
import os
import pathlib
import typing as t

import jinja2
from mcp import Tool
from pydantic import create_model

from nerve.models import Configuration
from nerve.tools.compiler import wrap_tool_function
from nerve.tools.mcp.client import Client


def _stringify_type(tp: t.Any) -> str:
    origin = t.get_origin(tp)
    args = t.get_args(tp)

    if origin is None:
        return tp.__name__ if hasattr(tp, "__name__") else str(tp)
    else:
        origin_name = origin.__name__ if hasattr(origin, "__name__") else str(origin)
        args_str = ", ".join(_stringify_type(arg) for arg in args)
        return f"{origin_name}[{args_str}]"


def _get_python_type(
    name: str, schema_item: dict[str, t.Any], dyn_defs: dict[str, t.Any] | None = None
) -> tuple[dict[str, t.Any] | None, t.Any]:
    if dyn_defs is None:
        dyn_defs = {}

    type_mapping = {
        "string": str,
        "integer": int,
        "number": float,
        "boolean": bool,
        "array": list,
        "object": dict,
        "null": None,
    }

    schema_type = schema_item.get("type")

    if schema_type == "array":
        items = schema_item.get("items", {})
        sub_dyn_defs, item_type = _get_python_type(name, items, dyn_defs)
        if sub_dyn_defs:
            dyn_defs.update(sub_dyn_defs)

        return dyn_defs, list[item_type]  # type: ignore
    elif schema_type == "object":
        properties = schema_item.get("properties", {})
        required = schema_item.get("required", [])

        field_definitions = {}
        for prop_name, prop_schema in properties.items():
            sub_dyn_defs, prop_type = _get_python_type(prop_name, prop_schema, dyn_defs)
            is_required = prop_name in required

            if sub_dyn_defs:
                dyn_defs.update(sub_dyn_defs)

            if is_required:
                field_definitions[prop_name] = (prop_type, ...)
            else:
                field_definitions[prop_name] = (prop_type | None, None)  # type: ignore

        # create a unique name for this model
        dyn_type_name = f"{name}_{len(dyn_defs)}"
        dyn_type = create_model(dyn_type_name, **field_definitions)  # type: ignore
        dyn_type.__module__ = None

        # store description metadata on the model if available
        for prop_name, prop_schema in properties.items():
            if "description" in prop_schema:
                setattr(dyn_type.__fields__[prop_name], "description", prop_schema["description"])

        dyn_defs[dyn_type.__name__] = dyn_type

        return dyn_defs, dyn_type
    else:
        return None, type_mapping.get(schema_type, t.Any)  # type: ignore


async def create_function_body(client: Client, mcp_tool: Tool) -> tuple[str, dict[str, t.Any]]:
    typed_args = []
    client_name = f"nerve_mcp_{client.name}_client"
    type_defs = {client_name: client}

    for name, arg_props in mcp_tool.inputSchema.get("properties", {}).items():
        # print(name, arg_props)
        args_def, arg_type = _get_python_type(name, arg_props)
        arg = {"name": name, "type": _stringify_type(arg_type), "description": arg_props.get("description", "")}

        if args_def:
            type_defs.update(args_def)

        if "default" in arg_props:
            arg["default"] = arg_props["default"]

        typed_args.append(arg)

    # load the template from the same directory as this script
    template_path = os.path.join(os.path.dirname(__file__), "body.j2")
    with open(template_path) as f:
        template_content = f.read()

    return (
        jinja2.Environment()
        .from_string(template_content)
        .render(client_name=client_name, tool=mcp_tool, arguments=typed_args),
        type_defs,
    )


async def get_tools_from_mcp(
    name: str, server: Configuration.MCPServer, working_dir: pathlib.Path
) -> list[t.Callable[..., t.Any]]:
    # connect and list tools
    client = Client(name, server, working_dir)
    mpc_tools = await client.tools()
    compiled_tools = []

    for mcp_tool in mpc_tools:
        func_body, type_defs = await create_function_body(client, mcp_tool)

        # print(func_body)
        exec(func_body, type_defs)

        tool_fn = wrap_tool_function(type_defs[mcp_tool.name])
        compiled_tools.append(tool_fn)
    return compiled_tools



import json
import pathlib
import time
import typing as t
from enum import Enum

import typer
from fastparquet import ParquetFile  # type: ignore[import-untyped]
from loguru import logger
from natsort import natsorted
from pydantic import BaseModel
from pydantic_yaml import parse_yaml_file_as

from nerve.runtime.runner import Arguments, Output


class Statistics(BaseModel):
    generator: str = ""
    max_steps: int = 0
    max_cost: float = 0
    timeout: float = 0
    window: str = ""
    runs: int = 0
    cases: int = 0
    passed: int = 0
    failed: int = 0


class Evaluation(BaseModel):
    name: str = ""
    started_at: float = 0.0
    finished_at: float = 0.0
    args: dict[str, t.Any] = {}
    runs: dict[str, list[Output]] = {}
    stats: Statistics = Statistics()

    _flush: bool = False

    @classmethod
    def build(cls, args: Arguments, runs: int, cases: int) -> "Evaluation":
        return Evaluation(
            started_at=time.time(),
            name=args.input_path.name,
            args=args.to_serializable(),
            stats=Statistics(
                generator=args.generator,
                max_steps=args.max_steps,
                max_cost=args.max_cost,
                timeout=args.timeout or 0.0,
                window=args.conversation_strategy_string,
                runs=runs,
                cases=cases,
            ),
        )

    def add_run(self, case_name: str, run_output: Output) -> None:
        if case_name not in self.runs:
            self.runs[case_name] = []

        self.runs[case_name].append(run_output)
        if run_output.task_success:
            self.stats.passed += 1
        else:
            self.stats.failed += 1

        self._flush = True

    def remove_run(self, case_name: str, run_idx: int) -> None:
        run_output = self.runs[case_name][run_idx]
        if run_output.steps > 0:
            # we're remove a completed run, take counters into account
            if run_output.task_success:
                self.stats.passed -= 1
            else:
                self.stats.failed -= 1

        self.runs[case_name].pop(run_idx)
        self._flush = True

    def num_runs(self, case_name: str) -> int:
        return len(self.runs[case_name]) if case_name in self.runs else 0

    def num_run_steps(self, case_name: str, run_idx: int) -> int:
        return self.runs[case_name][run_idx].steps

    def is_run_done(self, case_name: str, run_idx: int) -> bool:
        return self.num_run_steps(case_name, run_idx) > 0

    def get_run(self, case_name: str, run_idx: int) -> Output:
        return self.runs[case_name][run_idx]

    def needs_flush(self) -> bool:
        return self._flush

    def save_to(self, path: pathlib.Path) -> None:
        self.finished_at = time.time()
        path.write_text(self.model_dump_json())
        self._flush = False

    @classmethod
    def load_from(cls, path: pathlib.Path) -> "Evaluation":
        data = json.loads(path.read_text())
        return cls.model_validate(data)


class Case(BaseModel):
    name: str
    input_state: dict[str, t.Any]


class Cases:
    class Source(Enum):
        # cases have their own individual folders
        FOLDER = 0
        # cases are listed in a single file
        YAML = 1
        # parquet file
        PARQUET = 2

    def __init__(self, eval_path: pathlib.Path):
        self._eval_path = eval_path
        self._cases: list[Case] = []
        self._source = self.Source.FOLDER

        cases_folder = self._eval_path / "cases"
        cases_file_yml = self._eval_path / "cases.yml"
        cases_file_parquet = self._eval_path / "cases.parquet"

        if cases_folder.exists():
            self._from_folder(cases_folder)

        elif cases_file_yml.exists():
            self._from_yaml(cases_file_yml)

        elif cases_file_parquet.exists():
            self._from_parquet(cases_file_parquet)

        if not self._cases:
            logger.error(f"no cases found in {self._eval_path}")
            raise typer.Abort()

    def _from_folder(self, cases_folder: pathlib.Path) -> None:
        logger.info(f"📊 loading evaluation cases from folder {cases_folder}")
        self._source = self.Source.FOLDER
        for path in natsorted(cases_folder.glob("*")):
            self._cases.append(
                Case(
                    name=path.name,
                    input_state={
                        "CASE_NAME": path.name,
                        "CASE_PATH": path.absolute().as_posix(),
                    },
                )
            )

    def _from_yaml(self, cases_file: pathlib.Path) -> None:
        logger.info(f"📊 loading evaluation cases from file {cases_file}")
        self._source = self.Source.YAML
        for case in parse_yaml_file_as(list[dict[str, dict[str, t.Any]]], cases_file):  # type: ignore[type-var]
            for case_name, input_state in case.items():
                self._cases.append(Case(name=case_name, input_state=input_state))

    def _from_parquet(self, cases_file: pathlib.Path) -> None:
        logger.info(f"📊 loading evaluation cases from parquet file {cases_file}")
        self._source = self.Source.PARQUET
        pf = ParquetFile(cases_file)
        df = pf.to_pandas()
        num_rows = len(df)
        for index, row in df.iterrows():
            self._cases.append(
                Case(
                    name=f"case_{index}_of_{num_rows}",
                    input_state=row.to_dict(),
                )
            )

    def __iter__(self) -> t.Iterator[Case]:
        return iter(self._cases)

    def __len__(self) -> int:
        return len(self._cases)



import unittest
from typing import Annotated

from nerve.tools.protocol import get_tool_schema


class TestProtocol(unittest.IsolatedAsyncioTestCase):
    async def test_function_schema_creation(self) -> None:
        def _test_function(
            first: Annotated[str, "The first argument"],
            second: Annotated[int, "The second argument"] = 0,
        ) -> str | bytes:
            """Description."""
            return "hello"

        schema = get_tool_schema("", _test_function)
        self.assertEqual(
            schema,
            {
                "type": "function",
                "function": {
                    "name": "_test_function",
                    "description": "Description.",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "first": {"type": "string", "description": "The first argument"},
                            "second": {"type": "integer", "description": "The second argument"},
                        },
                        "required": ["first"],
                    },
                },
            },
        )



import pathlib
import tempfile
import unittest

from nerve.models import Configuration, Tool


class TestConfiguration(unittest.TestCase):
    def test_is_legacy_with_system_prompt(self) -> None:
        """Test that is_legacy returns True when system_prompt is set"""
        config = Configuration(system_prompt="This is a legacy prompt")
        self.assertTrue(config.is_legacy)

    def test_is_legacy_without_system_prompt(self) -> None:
        """Test that is_legacy returns False when system_prompt is None"""
        config = Configuration(agent="Agent prompt", task="Task description")
        self.assertFalse(config.is_legacy)

    def test_is_legacy_with_empty_config(self) -> None:
        """Test that is_legacy returns False with default configuration"""
        config = Configuration()
        self.assertFalse(config.is_legacy)

    def test_system_prompt_excluded_from_serialization(self) -> None:
        """Test that system_prompt is excluded when serializing the configuration"""
        config = Configuration(system_prompt="This should be excluded", agent="Agent prompt")
        serialized = config.model_dump()
        self.assertNotIn("system_prompt", serialized)

    def test_system_prompt_loaded_from_file(self) -> None:
        """Test that system_prompt is loaded if present in a file"""

        with tempfile.NamedTemporaryFile(suffix=".yml", delete=False) as temp_file:
            temp_file.write(b"system_prompt: Legacy prompt")
            temp_path = pathlib.Path(temp_file.name)

        try:
            loaded_config = Configuration.from_path(temp_path)

            self.assertEqual(loaded_config.system_prompt, "Legacy prompt")
            self.assertTrue(loaded_config.is_legacy)
        finally:
            temp_path.unlink()

    def test_get_inputs_from_agent_and_task(self) -> None:
        """Test that get_inputs extracts variables from both agent and task prompts"""
        config = Configuration(
            agent="I am an agent with {{input1}} and {{input2}}",
            task="Complete the task using {{input2}} and {{input3}}",
        )
        inputs = config.get_inputs()

        self.assertIn("input1", inputs)
        self.assertIn("input2", inputs)
        self.assertIn("input3", inputs)
        self.assertIsNone(inputs["input1"])
        self.assertIsNone(inputs["input2"])
        self.assertIsNone(inputs["input3"])

    def test_get_inputs_with_defaults(self) -> None:
        """Test that get_inputs includes default values when available"""
        config = Configuration(
            agent="I am an agent with {{input1}}",
            task="Complete the task with {{input2}}",
            defaults={"input1": "default1", "input2": "default2", "unused": "value"},
        )
        inputs = config.get_inputs()

        self.assertEqual(inputs["input1"], "default1")
        self.assertEqual(inputs["input2"], "default2")
        self.assertNotIn("unused", inputs)

    def test_get_inputs_from_tools(self) -> None:
        """Test that get_inputs extracts variables from tool definitions"""

        config = Configuration(
            agent="I am an agent",
            tools=[
                Tool(
                    name="test_tool",
                    tool="This tool uses {{global_var}} and {{tool_arg}}",
                    arguments=[Tool.Argument(name="tool_arg", description="A tool variable")],
                )
            ],
        )
        inputs = config.get_inputs()

        self.assertIn("global_var", inputs)
        self.assertNotIn("tool_arg", inputs)  # Should be excluded as it's a tool argument

    def test_get_inputs_from_all_sources(self) -> None:
        """Test that get_inputs extracts variables from all sources"""
        config = Configuration(
            agent="I am an agent with {{input1}}",
            task="Complete the task with {{input2}}",
            tools=[
                Tool(
                    name="test_tool",
                    tool="This tool uses {{global_var}} and {{tool_arg}}",
                    arguments=[Tool.Argument(name="tool_arg", description="A tool variable")],
                )
            ],
        )
        inputs = config.get_inputs()

        self.assertIn("input1", inputs)
        self.assertIn("input2", inputs)
        self.assertIn("global_var", inputs)
        self.assertNotIn("tool_arg", inputs)

    def test_get_inputs_includes_task_when_task_not_set(self) -> None:
        """Test that get_inputs includes 'task' when configuration.task is not set"""
        config = Configuration(
            agent="I am an agent with {{input1}}",
            task=None,
            defaults={"input1": "default1"},
        )
        inputs = config.get_inputs()

        self.assertIn("task", inputs)
        self.assertIn("input1", inputs)

    def test_get_inputs_ignores_interpolated_tool_calls(self) -> None:
        """Test that get_inputs doesn't extract variables from interpolated tool calls"""
        config = Configuration(
            agent="I am an agent",
            task="Complete the task using {{tool_name(param='value')}} and {{input1}}",
            tools=[
                Tool(
                    name="tool_name",
                    description="A test tool",
                )
            ],
        )
        inputs = config.get_inputs()

        self.assertIn("input1", inputs)
        self.assertNotIn("tool_name", inputs)  # Should be excluded as it's a tool
        self.assertNotIn("param", inputs)  # Should be excluded as it's a parameter to a tool call



import contextlib
import os
import pathlib
import sys
import typing as t

from loguru import logger

from nerve.models import Configuration, Tool
from nerve.tools import compiler
from nerve.tools.mcp import compiler as mcp_compiler


class Runtime:
    def __init__(self, name: str, generator: str, working_dir: pathlib.Path):
        # task unique identifier
        self.name = name
        # which model will handle this task
        self.generator = generator
        # the working directory for the task
        self.working_dir = working_dir
        # full chat history of the task
        self.history: list[t.Any] = []
        # available tools
        self.tools: list[t.Callable[..., t.Any]] = []

    @classmethod
    async def build(
        cls,
        working_dir: pathlib.Path,
        name: str,
        configuration: Configuration,
        debug: bool = False,
    ) -> "Runtime":
        logger.debug(
            f"building runtime for {name} with generator {configuration.generator} and working dir {working_dir}"
        )
        runtime = cls(name=name, generator=configuration.generator or "", working_dir=working_dir)

        # import tools from builtin namespaces
        ns_tools = compiler.get_tools_from_namespaces(configuration.using, configuration.jail)
        if ns_tools:
            logger.debug(f"🧰 importing {len(ns_tools)} tools from: {configuration.using}")
            runtime.tools.extend(ns_tools)

        # import custom tools from yaml definition
        yml_tools = compiler.get_tools_from_yml(
            working_dir,
            [tool for tool in configuration.tools if isinstance(tool, Tool) and not tool.path],
        )
        if yml_tools:
            logger.debug(f"🧰 importing {len(yml_tools)} tools from: {working_dir}")
            runtime.tools.extend(yml_tools)

        # import custom tools from files
        py_tools = compiler.get_tools_from_files(
            working_dir,
            [tool.path for tool in configuration.tools if isinstance(tool, Tool) and tool.path],
        )
        if py_tools:
            logger.debug(f"🧰 importing {len(py_tools)} tools from: {working_dir}")
            runtime.tools.extend(py_tools)

        # import custom tools from functions (when used as sdk)
        funcs = [compiler.wrap_tool_function(tool) for tool in configuration.tools if callable(tool)]
        if funcs:
            logger.debug(f"🧰 importing {len(funcs)} custom tools")
            runtime.tools.extend(funcs)

        # import MCP servers
        # tool initialization, especially for MCP servers, can be verbose,
        # we don't want to pollute the server logs with that
        out = open(os.devnull, "w") if not debug else sys.stdout
        err = open(os.devnull, "w") if not debug else sys.stderr
        for name, server in configuration.mcp.items():
            with contextlib.redirect_stdout(out), contextlib.redirect_stderr(err):
                server_tools = await mcp_compiler.get_tools_from_mcp(name, server, working_dir=working_dir)

            logger.info(f"🧰 importing {len(server_tools)} tools from MCP server {name}")
            runtime.tools.extend(server_tools)

        logger.debug(f"all runtime tools: {runtime.tools}")

        return runtime



import json
import pathlib
import typing as t

import typer
from loguru import logger

from nerve.defaults import (
    DEFAULT_AGENTS_LOAD_PATH,
    DEFAULT_CONVERSATION_STRATEGY,
    DEFAULT_GENERATOR,
    DEFAULT_MAX_COST,
    DEFAULT_MAX_STEPS,
    DEFAULT_TIMEOUT,
)
from nerve.generation import conversation
from nerve.runtime.runner import Arguments


def _resolve_input_path(input_path: pathlib.Path) -> pathlib.Path:
    # check if input_path exists
    if not input_path.exists():
        input_path_with_yaml = input_path.with_suffix(".yml")
        if input_path_with_yaml.exists():
            input_path = input_path_with_yaml

        elif not input_path.is_absolute():
            # check if it exists as part of the $HOME/.nerve/agents directory
            in_home = DEFAULT_AGENTS_LOAD_PATH / input_path
            if in_home.exists():
                input_path = in_home

            in_home_with_yaml = in_home.with_suffix(".yml")
            if in_home_with_yaml.exists():
                input_path = in_home_with_yaml

        if not input_path.exists():
            logger.error(f"path '{input_path}' does not exist")
            raise typer.Abort()

    return input_path


def _get_run_args(
    input_path: t.Annotated[
        pathlib.Path,
        typer.Argument(help="Agent or workflow to execute"),
    ] = pathlib.Path("."),
    task: t.Annotated[
        str | None,
        typer.Option("--task", "-t", help="Set or override the task for the agent."),
    ] = None,
    generator: t.Annotated[
        str,
        typer.Option("--generator", "-g", help="If the agent generator field is not set, use this generator."),
    ] = DEFAULT_GENERATOR,
    conversation_strategy: t.Annotated[
        str,
        typer.Option("--conversation", "-c", help="Conversation strategy to use"),
    ] = DEFAULT_CONVERSATION_STRATEGY,
    interactive: t.Annotated[
        bool,
        typer.Option("--interactive", "-i", help="Interactive mode"),
    ] = False,
    debug: t.Annotated[
        bool,
        typer.Option("--debug", help="Enable debug logging"),
    ] = False,
    litellm_debug: t.Annotated[
        bool,
        typer.Option("--litellm-debug", help="Enable litellm debug logging"),
    ] = False,
    litellm_tracing: t.Annotated[
        str | None,
        typer.Option("--litellm-tracing", help="Set litellm callbacks for tracing"),
    ] = None,
    quiet: t.Annotated[
        bool,
        typer.Option("--quiet", "-q", help="Quiet mode"),
    ] = False,
    max_steps: t.Annotated[
        int,
        typer.Option("--max-steps", "-s", help="Maximum number of steps. Set to 0 to disable."),
    ] = DEFAULT_MAX_STEPS,
    max_cost: t.Annotated[
        float,
        typer.Option(
            "--max-cost",
            help="If cost information is available, stop when the cost exceeds this value in USD. Set to 0 to disable.",
        ),
    ] = DEFAULT_MAX_COST,
    timeout: t.Annotated[
        int | None,
        typer.Option("--timeout", help="Timeout in seconds"),
    ] = DEFAULT_TIMEOUT,
    log_path: t.Annotated[
        pathlib.Path | None,
        typer.Option("--log", help="Log to a file."),
    ] = None,
    trace: t.Annotated[
        pathlib.Path | None,
        typer.Option("--trace", help="Save the final state to a file."),
    ] = None,
    start_state: t.Annotated[
        str,
        typer.Option("--start-state", help="Pass the initial input state as a JSON string."),
    ] = "{}",
) -> Arguments:
    return Arguments(
        input_path=_resolve_input_path(input_path),
        task=task,
        generator=generator,
        # convert the conversation strategy string to a valid enum
        conversation_strategy_string=conversation_strategy,
        conversation_strategy=conversation.strategy_from_string(conversation_strategy),
        interactive=interactive,
        debug=debug,
        litellm_debug=litellm_debug,
        litellm_tracing=litellm_tracing,
        quiet=quiet,
        max_steps=max_steps,
        max_cost=max_cost,
        timeout=timeout,
        log_path=log_path,
        trace=trace,
        # parse the start_state JSON string into a dictionary
        start_state=json.loads(start_state),
    )



import unittest
from unittest.mock import MagicMock

from mcp import Tool

from nerve.tools.mcp.client import Client
from nerve.tools.mcp.compiler import create_function_body


class TestCreateFunctionBody(unittest.IsolatedAsyncioTestCase):
    async def test_create_function_body_with_string_argument(self) -> None:
        # Mock the client
        client = MagicMock(spec=Client)

        # Create a mock tool with a simple string argument
        mock_tool = Tool(
            name="hello_world",
            description="A simple hello world function",
            inputSchema={
                "type": "object",
                "properties": {"name": {"type": "string", "description": "The name to greet"}},
                "required": ["name"],
            },
        )

        func_body, type_defs = await create_function_body(client, mock_tool)

        self.assertIn(
            '''
async def hello_world(name: Annotated[str, "The name to greet"]) -> Any:
    """A simple hello world function"""
'''.strip(),
            func_body,
        )

    async def test_create_function_body_with_string_and_int_arguments(self) -> None:
        # Mock the client
        client = MagicMock(spec=Client)

        # Create a mock tool with string and int arguments
        mock_tool = Tool(
            name="calculate",
            description="A function that performs a calculation",
            inputSchema={
                "type": "object",
                "properties": {
                    "operation": {"type": "string", "description": "The operation to perform"},
                    "value": {"type": "integer", "description": "The value to calculate with"},
                },
                "required": ["operation", "value"],
            },
        )

        func_body, type_defs = await create_function_body(client, mock_tool)

        self.assertIn(
            '''
async def calculate(operation: Annotated[str, "The operation to perform"], value: Annotated[int, "The value to calculate with"]) -> Any:
    """A function that performs a calculation"""
'''.strip(),
            func_body,
        )

    async def test_create_function_body_with_string_and_int_arguments_with_default(self) -> None:
        # Mock the client
        client = MagicMock(spec=Client)

        # Create a mock tool with string and int arguments
        mock_tool = Tool(
            name="calculate",
            description="A function that performs a calculation",
            inputSchema={
                "type": "object",
                "properties": {
                    "operation": {"type": "string", "description": "The operation to perform"},
                    "value": {"type": "integer", "description": "The value to calculate with", "default": 10},
                },
                "required": ["operation"],
            },
        )

        func_body, type_defs = await create_function_body(client, mock_tool)

        self.assertIn(
            '''
async def calculate(operation: Annotated[str, "The operation to perform"], value: Annotated[int, "The value to calculate with"] = 10) -> Any:
    """A function that performs a calculation"""
'''.strip(),
            func_body,
        )

    async def test_create_function_body_with_complex_nested_arguments(self) -> None:
        # Mock the client
        client = MagicMock(spec=Client)

        # Create a mock tool with complex nested arguments
        mock_tool = Tool(
            name="process_data",
            description="A function that processes complex data structures",
            inputSchema={
                "type": "object",
                "properties": {
                    "user": {
                        "type": "object",
                        "properties": {
                            "name": {"type": "string", "description": "User's name"},
                            "age": {"type": "integer", "description": "User's age"},
                            "preferences": {
                                "type": "array",
                                "items": {"type": "string"},
                                "description": "User's preferences",
                            },
                        },
                        "description": "User information",
                    },
                    "settings": {
                        "type": "object",
                        "properties": {
                            "enabled": {"type": "boolean", "description": "Whether processing is enabled"},
                            "options": {
                                "type": "object",
                                "properties": {
                                    "mode": {"type": "string", "description": "Processing mode"},
                                    "priority": {"type": "integer", "description": "Processing priority"},
                                },
                                "description": "Processing options",
                            },
                        },
                        "description": "Processing settings",
                    },
                },
                "required": ["user", "settings"],
            },
        )

        func_body, type_defs = await create_function_body(client, mock_tool)

        print()
        print(func_body)
        print()

        self.assertIn(
            '''
async def process_data(user: Annotated[user_0, "User information"], settings: Annotated[settings_1, "Processing settings"]) -> Any:
    """A function that processes complex data structures"""
'''.strip(),
            func_body,
        )



import asyncio
import pathlib
import typing as t

import typer
from loguru import logger
from termcolor import colored
from typer_di import Depends, TyperDI

import nerve
from nerve.cli.utils import _get_run_args
from nerve.defaults import DEFAULT_EVAL_RUNS
from nerve.models import Configuration
from nerve.runtime import logging
from nerve.runtime.eval import Case, Cases, Evaluation
from nerve.runtime.runner import Arguments, Output, Runner

cli = TyperDI(
    no_args_is_help=True,
    pretty_exceptions_enable=False,
    context_settings={"help_option_names": ["-h", "--help"]},
)


def _get_output_path(args: Arguments) -> pathlib.Path:
    output_name = f"{args.generator}-{args.input_path.name}"
    sanitized = ""
    for char in output_name:
        if char.isalnum() or char in "._- ":
            sanitized += char
        else:
            sanitized += "-"

    while "--" in sanitized:
        sanitized = sanitized.replace("--", "-")

    return pathlib.Path(f"{sanitized}.json")


@cli.command(
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True, "help_option_names": ["-h", "--help"]},
    help="Execute an agent or a workflow in evaluation mode.",
)
def eval(
    args: Arguments = Depends(_get_run_args),
    runs: t.Annotated[int, typer.Option("--runs", "-r", help="Number of runs per case.")] = DEFAULT_EVAL_RUNS,
    output: t.Annotated[
        pathlib.Path | None,
        typer.Option("--output", "-o", help="Path to save the output. If not set will be auto generated."),
    ] = None,
) -> None:
    logging.init(
        args.log_path,
        level="DEBUG" if args.debug else "SUCCESS" if args.quiet else "INFO",
        litellm_debug=args.litellm_debug,
    )
    logger.info(f"🧠 nerve v{nerve.__version__}")

    try:
        config = Configuration.from_path(args.input_path)
    except Exception as e:
        logger.error(f"path '{args.input_path}' is not a valid agent configuration: {e}")
        raise typer.Abort() from e

    output = output or _get_output_path(args)
    cases = Cases(args.input_path)

    # apply limits from the config if available
    if config.limits:
        if config.limits.runs:
            runs = config.limits.runs
        if config.limits.max_steps:
            args.max_steps = config.limits.max_steps
        if config.limits.max_cost:
            args.max_cost = config.limits.max_cost
        if config.limits.timeout:
            args.timeout = config.limits.timeout

    eval_name = colored(args.input_path.name, "green", attrs=["bold"])
    logger.info(f"📊 {args.generator} / {eval_name} / cases: {len(cases)} / runs: {runs}")

    if output.exists():
        logger.info(f"📊 loading evaluation results from {output}")
        evaluation = Evaluation.load_from(output)
    else:
        logger.info(f"📊 saving evaluation results to {output}")
        evaluation = Evaluation.build(args, runs, len(cases))

    for case in cases:
        for run in range(runs):
            do_run = True
            if evaluation.num_runs(case.name) >= runs:
                # we already have enough runs for this case
                do_run = False
                if not evaluation.is_run_done(case.name, run):
                    # we don't have enough runs for this case
                    do_run = True
                    logger.warning(f"run {run} for {case.name} has not been completed, re-running")
                    evaluation.remove_run(case.name, run)

            if not do_run:
                logger.debug(f"skipping {case.name} ({run + 1}/{runs})")
                run_output = evaluation.get_run(case.name, run)
            else:
                logger.debug(f"running {case.name} ({run + 1}/{runs})")
                run_output = asyncio.run(_run_case(args, case))
                evaluation.add_run(case.name, run_output)

            _show_run(args, run_output, runs, run, case.name, do_run)

            if evaluation.needs_flush():
                # save at each run so we can restore later
                evaluation.save_to(output)

    logger.debug(f"evaluation results: {evaluation}")

    # save if needed
    if evaluation.needs_flush():
        evaluation.save_to(output)
        logger.info(f"📊 evaluation results saved to {output}")

    _show_results(evaluation)


def _show_run(args: Arguments, output: Output, runs: int, run: int, case_name: str, live: bool) -> None:
    usage = output.usage
    one_of = f"[{run + 1}/{runs}]" if live else f"({run + 1}/{runs})"
    subject = f"{one_of} {args.generator} / {args.input_path.name} / {case_name}"
    stats = (
        f"{output.steps} steps, {output.time:.1f} s, {usage.get('total_tokens', 0)} tokens, {usage.get('cost', 0.0)} $"
    )
    if output.task_success:
        logger.success(f"   {subject} : {stats}")
    else:
        logger.error(f"     {subject} : {stats}")


def _show_results(eval: Evaluation) -> None:
    print()
    logger.info("📊 Results")
    logger.info(f"Model: {eval.args['generator']}")
    logger.info(f"Cases: {eval.stats.cases}")
    logger.info(f"Runs: {eval.stats.runs}")
    logger.info(f"Pass: {eval.stats.passed}")
    logger.info(f"Fail: {eval.stats.failed}")

    total_cost = 0.0
    # total_tokens = 0
    total_steps = 0
    total_time = 0.0
    total_tests = eval.stats.passed + eval.stats.failed
    score = eval.stats.passed / total_tests * 100

    for _case_name, case_runs in eval.runs.items():
        for run in case_runs:
            total_cost += run.usage.get("cost", 0.0)
            # total_tokens += run.usage.get("total_tokens", 0)
            total_steps += run.steps
            total_time += run.time

    logger.info(f"Total cost: {total_cost:.2f} $")
    logger.info(f"Total time: {total_time:.2f} s")
    logger.info(f"Avg time: {total_time / total_tests:.2f} s")
    logger.info(f"Avg steps: {total_steps / total_tests:.2f}")
    logger.info(f"Avg cost: {total_cost / total_tests} $")

    logger.info(f"Score: {score:.2f} %")


async def _run_case(args: Arguments, case: Case) -> Output:
    return await Runner(
        args,
        case.input_state,
    ).run()



import asyncio
import os
import pathlib
import sys
import typing as t
from contextlib import AsyncExitStack, asynccontextmanager

import mcp.types as mcp_types
from loguru import logger
from mcp import ClientSession, StdioServerParameters, Tool
from mcp.client.sse import sse_client
from mcp.client.stdio import stdio_client
from mcp.types import EmbeddedResource, ImageContent, TextContent

from nerve.models import Configuration


class Client:
    name: str = ""

    def __init__(self, name: str, server: Configuration.MCPServer, working_dir: pathlib.Path):
        for key, value in server.env.items():
            if not value:
                env_val = os.getenv(key)
                if env_val:
                    logger.debug("setting {} from env", key)
                    server.env[key] = env_val

            if not server.env[key]:
                logger.error("mcp server {} environment variable {} is not set", self.name, key)
                exit(1)

        # pass all env
        server.env.update(os.environ)

        self.name = name
        self.server = server
        self.server_params = StdioServerParameters(
            command=server.command, args=server.args, env=server.env, cwd=working_dir
        )
        self._session: ClientSession | None = None
        self._exit_stack = AsyncExitStack()
        self._tools: list[Tool] = []
        self._client_context: t.Any = None

    async def _logging_callback(
        self,
        params: mcp_types.LoggingMessageNotificationParams,
    ) -> None:
        line = str(params.data)
        # parts = line.split("]", 2)
        # line = parts[1].strip() if len(parts) > 1 else line  # remove timestamp
        # line = line.replace(str(params.level).upper(), "").strip()  # remove level
        getattr(logger, str(params.level))(f"<{self.name}> {line}")

    @asynccontextmanager
    async def _safe_stdio_context(
        self,
    ) -> t.AsyncGenerator[tuple[t.Any, t.Any], None]:
        gen_exit = False
        try:
            async with stdio_client(server=self.server_params, errlog=sys.stderr) as (read_stream, write_stream):
                logger.debug("stdio streams for {} created", self.name)
                try:
                    yield read_stream, write_stream
                except GeneratorExit:
                    logger.debug("GeneratorExit")
                    gen_exit = True
        except Exception:
            logger.debug(f"Exception, gen_exit = {gen_exit}")
            # TODO: there's a weird bug, if we don't do this when the process exits
            # we will see an exception
            if gen_exit:
                exit(0)

    async def connect(self) -> None:
        if self._session:
            return

        self._client_context = None
        if self.server.url:
            logger.debug("connecting to SSE MCP server {}", self.name)
            self._client_context = sse_client(
                url=self.server.url or "http://localhost:8080",
                headers=self.server.headers,
                timeout=self.server.timeout,
                sse_read_timeout=self.server.sse_read_timeout,
            )
        else:
            logger.debug("connecting to stdio MCP server {}", self.name)
            self._client_context = self._safe_stdio_context()

        self._read_stream, self._write_stream = await self._exit_stack.enter_async_context(self._client_context)
        logger.debug("creating async context for {}", self.name)
        self._session = await self._exit_stack.enter_async_context(
            ClientSession(
                read_stream=self._read_stream, write_stream=self._write_stream, logging_callback=self._logging_callback
            )
        )

        try:
            logger.debug("initializing session for {}", self.name)
            await asyncio.wait_for(self._session.initialize(), timeout=self.server.session_timeout)
        except asyncio.TimeoutError:
            logger.error(
                "mcp server {} initialization timeout",
                self.name,
            )
            exit(1)

        logger.debug("session initialized for {}", self.name)

        self._tools = await self.tools()

        logger.debug("connected to MCP server {} with {} tools", self.name, len(self._tools))
        for tool in self._tools:
            logger.debug("tool: {}", tool)

    async def tools(self) -> list[Tool]:
        if not self._session:
            await self.connect()
            if not self._session:
                raise Exception("failed to connect to MCP server")

        if self._tools:
            return self._tools

        logger.debug("listing tools from MCP server {}", self.name)

        response = await self._session.list_tools()
        logger.debug("list_tools.response: {}", response)
        self._tools = response.tools

        return self._tools

    async def call_tool(self, mcp_tool_name: str, **kwargs: t.Any) -> t.Any:
        if not self._session:
            await self.connect()
            if not self._session:
                raise Exception("failed to connect to MCP server")

        logger.debug("calling mcp tool {} with args: {}", mcp_tool_name, kwargs)
        ret = await self._session.call_tool(mcp_tool_name, kwargs)
        logger.debug("mcp {} call result: {}", mcp_tool_name, ret)

        if ret.isError:
            raise Exception(str(ret))

        responses: list[t.Any] = []
        for elem in ret.content:
            if isinstance(elem, TextContent):
                responses.append(elem.text)
            elif isinstance(elem, ImageContent):
                responses.append(
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:{elem.mimeType};base64,{elem.data}"},
                    }
                )
            elif isinstance(elem, EmbeddedResource):
                raise Exception("EmbeddedResource not supported yet")

        logger.debug("tool call responses: {}", responses)

        return responses[0] if len(responses) == 1 else responses



import pathlib
from unittest.mock import MagicMock, patch

import pytest
import typer

from nerve.cli.utils import _resolve_input_path
from nerve.defaults import DEFAULT_AGENTS_LOAD_PATH


def test_resolve_input_path_exists() -> None:
    """Test when input_path exists directly."""
    mock_path = MagicMock(spec=pathlib.Path)
    mock_path.exists.return_value = True

    result = _resolve_input_path(mock_path)

    assert result == mock_path
    mock_path.exists.assert_called_once()


def test_resolve_input_path_with_yaml_extension() -> None:
    """Test when input_path with .yml extension exists."""
    mock_path = MagicMock(spec=pathlib.Path)
    mock_path.exists.side_effect = [False, True]
    mock_path_with_yaml = MagicMock(spec=pathlib.Path)
    mock_path.with_suffix.return_value = mock_path_with_yaml
    mock_path_with_yaml.exists.return_value = True

    result = _resolve_input_path(mock_path)

    assert result == mock_path_with_yaml
    mock_path.with_suffix.assert_called_once_with(".yml")


def test_resolve_input_path_in_home_with_yaml() -> None:
    """Test when input_path with .yml extension exists in the default agents directory."""
    mock_path = MagicMock(spec=pathlib.Path)
    mock_path.exists.side_effect = [False, False]
    mock_path.with_suffix.return_value.exists.return_value = False
    mock_path.is_absolute.return_value = False

    in_home_path = MagicMock(spec=pathlib.Path)
    in_home_path.exists.return_value = False

    in_home_yaml_path = MagicMock(spec=pathlib.Path)
    in_home_yaml_path.exists.return_value = True
    in_home_path.with_suffix.return_value = in_home_yaml_path

    with patch("nerve.cli.utils.DEFAULT_AGENTS_LOAD_PATH", DEFAULT_AGENTS_LOAD_PATH):
        with patch("pathlib.Path.__truediv__", return_value=in_home_path):
            result = _resolve_input_path(mock_path)

    assert result == in_home_yaml_path
    in_home_path.with_suffix.assert_called_once_with(".yml")


def test_resolve_input_path_not_exists() -> None:
    """Test when input_path does not exist anywhere, should raise typer.Abort."""
    mock_path = MagicMock(spec=pathlib.Path)
    mock_path.exists.side_effect = [False, False]
    mock_path.with_suffix.return_value.exists.return_value = False
    mock_path.is_absolute.return_value = False

    in_home_path = MagicMock(spec=pathlib.Path)
    in_home_path.exists.return_value = False
    in_home_path.with_suffix.return_value.exists.return_value = False

    with patch("nerve.cli.utils.DEFAULT_AGENTS_LOAD_PATH", DEFAULT_AGENTS_LOAD_PATH):
        with patch("pathlib.Path.__truediv__", return_value=in_home_path):
            with patch("nerve.cli.utils.logger.error") as mock_logger:
                with pytest.raises(typer.Abort):
                    _resolve_input_path(mock_path)

    mock_logger.assert_called_once()
